/*
**	 linux/amiga/amicon.c
**
**	 Copyright (C) 1994 David Carter  [carter@compsci.bristol.ac.uk]
**		  Modified 20-Apr-94 by William Rucklidge (wjr@cs.cornell.edu)
**				- AGA modes
**		  Modified 03-May-94 by Geert Uytterhoeven
**									  (Geert.Uytterhoeven@cs.kuleuven.ac.be)
**				- additional AGA modes
**				- 24-bit color palette
**		  Modified 19-June-94 by Geert Uytterhoeven
**				- cursor flashing for better cursor visibility
**		  Modified 24-Jan-95 by Geert Uytterhoeven
**				- frame buffer update
**		  Modified 18-August-94 by Jes Sorensen (jds@kom.auc.dk)
**				- ECS support. The following screen-modes are included:
**					 multiscan, multiscan-lace, super72, super72-lace, dblntsc,
**					 dblpal & euro72.
**				  I suggest that we remove the old AGA screenmodes, as they are
**					 non-standard, and some of them doesn't work under ECS.
**		  Minor cleanup on 15-November-94, first release of the ECS-screenmodes.
**				  (Jes)
**		  Modified Aug-95 by Geert Uytterhoeven
**				- many changes to support the new frame buffer device (color,
**				  interleaved/normal bitplanes, ...)
**
**	 Interface based on the original `amicon.c' code:
**		  Copyright (C) 1993 Hamish Macdonald and Greg Harp
**	 although the contents have been radically altered.
**
**  This file is subject to the terms and conditions of the GNU General
**  Public License.  See the file README.legal in the main directory of this
**  archive for more details.
*/


#undef DYNAMIC_RESIZE	/* ++Geert: Doesn't work yet due to kmalloc() */


#include <linux/types.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>

#include <linux/amigahw.h>
#include <linux/config.h>				/* New: DPC */
#include <linux/bootinfo.h>
#include <linux/sched.h>
#include <linux/interrupt.h>			/* New: DPC */
#include <linux/amigaints.h>
#include <linux/kd.h>

#include <asm/system.h>

#include "amifb.h"

#include "../drivers/char/vt_kern.h"	/* ++Geert: for vt_cons */


struct fb_info *fb_info;
struct display *disp;


/* ++Geert: Sorry, no hardware cursor support at the moment;
	use Atari alike software cursor */

static int cursor_drawn = 0;

#define CURSOR_DRAW_DELAY				2

/* # VBL ints between cursor state changes */
#define DEFAULT_CURSOR_BLINK_RATE	32

static int vbl_cursor_cnt = 0;
static int cursor_on = 0;
static int cursor_blink_rate = DEFAULT_CURSOR_BLINK_RATE;

static __inline__ int CURSOR_UNDRAWN(void)
{
	int cursor_was_drawn;
	vbl_cursor_cnt = 0;
	cursor_was_drawn = cursor_drawn;
	cursor_drawn = 0;
	return(cursor_was_drawn);
}

	/*
	 *		Attribute Decoding
	 */

#define attr_fgcol(attr)		((attr) & 0x0f)			/* Color */
#define attr_bgcol(attr)		(((attr)>>4) & 0x0f)

#define attr_bold(attr)			((attr & 3) == 2)			/* Monochrome */
#define attr_reverse(attr)		((attr) & 8)
#define attr_underline(attr)	((attr) & 4)


	/*
	 *		Interface used by the world
	 */

static int amicon_changevar(int con);
static long amicon_init(struct condata *conp, long kmem_start,
								char **display_desc);
static int amicon_deinit(struct condata *conp);
static int amicon_clear(struct condata *conp, int sy, int sx, int height,
								int width);
static int amicon_putc(struct condata *conp, int c, int y, int x);
static int amicon_putcs(struct condata *conp, const char *s, int count, int y,
								int x);
static int amicon_cursor(struct condata *conp, int mode);
static int amicon_scroll(struct condata *conp, int t, int b, int dir,
								 int count);
static int amicon_bmove(struct condata *conp, int sy, int sx, int dy, int dx,
								int height, int width);
static int amicon_switch(struct condata *conp);
static int amicon_blank(int blank);


	/*
	 *		Imported from drivers/char/console.c
	 */

extern void con_clear(int currcons);
#ifdef DYNAMIC_RESIZE
extern void con_changeprop(int currcons, int oldcols, int oldrows);
#endif /* DYNAMIC_RESIZE */


	/*
	 *		Internal routines
	 */

static void amicon_setup(int con, int setcol, int cls);
static __inline__ void *mymemclear_small(void *s, size_t count);
static __inline__ void *mymemclear(void *s, size_t count);
static __inline__ void *mymemset(void *s, size_t count);
static __inline__ void *mymemmove(void *d, void *s, size_t count);
static __inline__ int real_y(struct display *p, int y);
static void amicon_vbl_handler(struct intframe *fp, void *dummy);
static void amicon_bmove_rec(struct display *p, int sy, int sx, int dy, int dx,
									  int height, int width, u_int y_break);


	/*
	 *		Support for Monochrome
	 */

static void amicon_bmove_mono(struct display *p, int sy, int sx, int dy, int dx,
										int height, int width);
static void amicon_clear_mono(struct condata *conp, struct display *p, int sy,
										int sx, int height, int width);
static void amicon_putc_mono(struct condata *conp, struct display *p, int c,
									  int y, int x);
static void amicon_putcs_mono(struct condata *conp, struct display *p,
										const char *s, int count, int y, int x);
static void amicon_rev_char_mono(struct display *p, int x, int y);


	/*
	 *		Support for Color Interleaved Planes
	 */

static void amicon_bmove_ilbm(struct display *p, int sy, int sx, int dy, int dx,
										int height, int width);
static void amicon_clear_ilbm(struct condata *conp, struct display *p, int sy,
										int sx, int height, int width);
static void amicon_putc_ilbm(struct condata *conp, struct display *p, int c,
									  int y, int x);
static void amicon_putcs_ilbm(struct condata *conp, struct display *p,
										const char *s, int count, int y, int x);
static void amicon_rev_char_ilbm(struct display *p, int x, int y);


	/*
	 *		Support for Color Planes
	 */

static void amicon_bmove_plan(struct display *p, int sy, int sx, int dy, int dx,
										int height, int width);
static void amicon_clear_plan(struct condata *conp, struct display *p, int sy,
										  int sx, int height, int width);
static void amicon_putc_plan(struct condata *conp, struct display *p, int c,
										 int y, int x);
static void amicon_putcs_plan(struct condata *conp, struct display *p,
										  const char *s, int count, int y, int x);
static void amicon_rev_char_plan(struct display *p, int x, int y);


	/*
	 *		`switch' for the Low Level Operations
	 */

struct display_switch {
    void (*bmove)(register struct display *p, int sy, int sx, int dy, int dx,
						int height, int width);
    void (*clear)(struct condata *conp, register struct display *p, int sy,
						int sx, int height, int width);
    void (*putc)(struct condata *conp, struct display *p, int c, int y, int x);
    void (*putcs)(struct condata *conp, struct display *p, const char *s,
						int count, int y, int x);
    void (*rev_char)(struct display *display, int x, int y);
};


struct display_switch dispsw_mono = {
	amicon_bmove_mono, amicon_clear_mono, amicon_putc_mono, amicon_putcs_mono,
	amicon_rev_char_mono
};

struct display_switch dispsw_ilbm = {
	amicon_bmove_ilbm, amicon_clear_ilbm, amicon_putc_ilbm, amicon_putcs_ilbm,
	amicon_rev_char_ilbm
};

struct display_switch dispsw_plan = {
	amicon_bmove_plan, amicon_clear_plan, amicon_putc_plan, amicon_putcs_plan,
	amicon_rev_char_plan
};


static void amicon_setup(int con, int setcol, int cls)
{
	struct display *p = &disp[con];
	struct condata *conp = &vc_cons[con];
#ifdef DYNAMIC_RESIZE
	int oldcols = conp->vc_cols;
	int oldrows = conp->vc_rows;
#endif /* DYNAMIC_RESIZE */

	p->var.yoffset = 0;	/* reset ywrap */

	conp->vc_cols = p->var.xres/p->fontwidth;
	conp->vc_rows = p->var.yres/p->fontheight;
	conp->vc_can_do_color = (p->var.bits_per_pixel != 1);

	switch (p->type) {
		case FB_TYPE_INTERLEAVED_PLANES:
      	p->next_line = p->type_aux;
      	p->next_plane = p->type_aux/p->var.bits_per_pixel;
			if (conp->vc_can_do_color)
				p->dispsw = &dispsw_ilbm;
			else
				p->dispsw = &dispsw_mono;
			break;

		case FB_TYPE_PLANES:
      	p->next_line = (p->var.xres_virtual)>>3;
      	p->next_plane = p->var.yres_virtual*p->next_line;
			if (conp->vc_can_do_color)
				p->dispsw = &dispsw_plan;
			else
				p->dispsw = &dispsw_mono;
			break;

		case FB_TYPE_PACKED_PIXELS:
      	p->next_line = (p->var.xres_virtual*p->var.bits_per_pixel)>>3;
      	p->next_plane = 0;
			if (conp->vc_can_do_color)
	  			goto fail;
	  		else
				p->dispsw = &dispsw_mono;

		default:
fail:		printk("amicon_setup: type %d (aux %d) not supported, trying mono\n",
					 p->type, p->type_aux);
			p->next_line = (p->var.xres_virtual)>>3;
      	p->next_plane = 0;
			p->var.bits_per_pixel = 1;
			p->dispsw = &dispsw_mono;
			break;
	}

	if (setcol) {
		p->fgcol = p->var.bits_per_pixel > 2 ? 7 : 1;
		p->bgcol = 0;
	}

	if (cls) {
#ifndef DYNAMIC_RESIZE
		/* clear screen (a dirty hack...) */
		con_clear(con);
		/* ++guenther: once again, because attributes are updated only after
		 * clearing in con_clear() (with update_attr(), which is static). 
		 * This causes trouble with switching from color to monochr. console.
		 * It would be better, if attributes were updated just before the
		 * console is actually cleared (in drivers/char/console.c), but
		 * this may have other disadvantages, which I don't see yet.
		 */
		con_clear(con);
#else /* DYNAMIC_RESIZE */
		con_changeprop(con, oldcols, oldrows);
#endif /* DYNAMIC_RESIZE */
	}

	conp->vc_flags |= CON_INITED;
}


static int amicon_changevar(int con)
{
	amicon_setup(con, 1, 1);
	return(0);
}


static long amicon_init(struct condata *conp, long kmem_start,
								char**display_desc)
{
	int unit = conp-vc_cons;
	static int master_init = 0;

	if (!master_init) {
		master_init = 1;
		fb_info = amifb_init(&kmem_start);
		disp = fb_info->disp;

		*display_desc = fb_info->modename;
		fb_info->changevar = &amicon_changevar;

		add_isr(IRQ_AMIGA_VERTB, amicon_vbl_handler, 0, NULL);
	} else {
		/* Copy all from the first console, maybe allow different
			display modes later. */
		disp[unit] = disp[0];
	}
	amicon_setup(unit, 1, 0);

	return(kmem_start);
}


static int amicon_deinit(struct condata *conp)
{
	struct display *p = &disp[conp-vc_cons];

	/* You can't deinit a non inited screen nor the console 0. */
	if (!(conp->vc_flags & CON_INITED) || (p == disp))
		return(-1);

	amicon_switch(&vc_cons[0]);	/* Always switch to the console 0 */
	memset(p, 0, sizeof(struct display));
	conp->vc_flags &= ~CON_INITED;

	return(0);
}


/* ================================================================= */
/*                      Utility Assembler Functions                  */
/* ================================================================= */


/* ====================================================================== */

/* Those of a delicate disposition might like to skip the next couple of
 * pages.
 *
 * These functions are drop in replacements for memmove and
 * memset(_, 0, _). However their five instances add at least a kilobyte
 * to the object file. You have been warned.
 *
 * Not a great fan of assembler for the sake of it, but I think
 * that these routines are at least 10 times faster than their C
 * equivalents for large blits, and thats important to the lowest level of
 * a graphics driver. Question is whether some scheme with the blitter
 * would be faster. I suspect not for simple text system - not much
 * asynchronisity.
 *
 * Code is very simple, just gruesome expansion. Basic strategy is to
 * increase data moved/cleared at each step to 16 bytes to reduce
 * instruction per data move overhead. movem might be faster still
 * For more than 15 bytes, we try to align the write direction on a
 * longword boundary to get maximum speed. This is even more gruesome.
 * Unaligned read/write used requires 68020+ - think this is a problem?
 *
 * Sorry!
 */


/* ++roman: I've optimized Robert's original versions in some minor
 * aspects, e.g. moveq instead of movel, let gcc choose the registers,
 * use movem in some places...
 * For other modes than 1 plane, lots of more such assembler functions
 * were needed (e.g. the ones using movep or expanding color values).
 */

/* ++andreas: more optimizations:
   subl #65536,d0 replaced by clrw d0; subql #1,d0 for dbcc
   addal is faster than addaw
   movep is rather expensive compared to ordinary move's
   some functions rewritten in C for clearity, no speed loss */

static __inline__ void *mymemclear_small(void *s, size_t count)
{
	if (!count)
		return(0);

	__asm__ __volatile__(
			"lsrl   #1,%1 ; jcc 1f ; moveb %2,%0@-\n\t"
		"1: lsrl   #1,%1 ; jcc 1f ; movew %2,%0@-\n\t"
		"1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@-\n\t"
		"1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@- ; movel %2,%0@-\n\t"
		"1: subql  #1,%1 ; jcs 3f\n\t"
		"2: moveml %2/%3/%4/%5,%0@-\n\t"
			"dbra %1,2b\n\t"
		"3:"
			: "=a" (s), "=d" (count)
			:  "d" (0), "d" (0), "d" (0), "d" (0),
				"0" ((char *)s+count), "1" (count)
	);

	return(0);
}


static __inline__ void *mymemclear(void *s, size_t count)
{
	if (!count)
		return(0);

	if (count < 16) {
		__asm__ __volatile__(
				"lsrl   #1,%1 ; jcc 1f ; clrb %0@+\n\t"
			"1: lsrl   #1,%1 ; jcc 1f ; clrw %0@+\n\t"
			"1: lsrl   #1,%1 ; jcc 1f ; clrl %0@+\n\t"
			"1: lsrl   #1,%1 ; jcc 1f ; clrl %0@+ ; clrl %0@+\n\t"
			"1:"
				: "=a" (s), "=d" (count)
				: "0" (s), "1" (count)
		);
	} else {
		long tmp;
		__asm__ __volatile__(
				"movel %1,%2\n\t"
				"lsrl   #1,%2 ; jcc 1f ; clrb %0@+ ; subqw #1,%1\n\t"
				"lsrl   #1,%2 ; jcs 2f\n\t"  /* %0 increased=>bit 2 switched*/
				"clrw   %0@+  ; subqw  #2,%1 ; jra 2f\n\t"
			"1: lsrl   #1,%2 ; jcc 2f\n\t"
				"clrw   %0@+  ; subqw  #2,%1\n\t"
			"2: movew %1,%2; lsrl #2,%1 ; jeq 6f\n\t"
				"lsrl   #1,%1 ; jcc 3f ; clrl %0@+\n\t"
			"3: lsrl   #1,%1 ; jcc 4f ; clrl %0@+ ; clrl %0@+\n\t"
			"4: subql  #1,%1 ; jcs 6f\n\t"
			"5: clrl %0@+; clrl %0@+ ; clrl %0@+ ; clrl %0@+\n\t"
				"dbra %1,5b   ; clrw %1; subql #1,%1; jcc 5b\n\t"
			"6: movew %2,%1; btst #1,%1 ; jeq 7f ; clrw %0@+\n\t"
			"7:            ; btst #0,%1 ; jeq 8f ; clrb %0@+\n\t"
			"8:"
				: "=a" (s), "=d" (count), "=d" (tmp)
				: "0" (s), "1" (count)
		);
	}

	return(0);
}


static __inline__ void *mymemset(void *s, size_t count)
{
	if (!count)
		return(0);

	__asm__ __volatile__(
			"lsrl   #1,%1 ; jcc 1f ; moveb %2,%0@-\n\t"
		"1: lsrl   #1,%1 ; jcc 1f ; movew %2,%0@-\n\t"
		"1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@-\n\t"
		"1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@- ; movel %2,%0@-\n\t"
		"1: subql  #1,%1 ; jcs 3f\n\t"
		"2: moveml %2/%3/%4/%5,%0@-\n\t"
			"dbra %1,2b\n\t"
		"3:"
			: "=a" (s), "=d" (count)
			:  "d" (-1), "d" (-1), "d" (-1), "d" (-1),
				"0" ((char *) s + count), "1" (count)
	);

	return(0);
}


static __inline__ void *mymemmove(void *d, void *s, size_t count)
{
	if (d < s) {
		if (count < 16) {
			__asm__ __volatile__(
					"lsrl   #1,%2 ; jcc 1f ; moveb %1@+,%0@+\n\t"
				"1: lsrl   #1,%2 ; jcc 1f ; movew %1@+,%0@+\n\t"
				"1: lsrl   #1,%2 ; jcc 1f ; movel %1@+,%0@+\n\t"
				"1: lsrl   #1,%2 ; jcc 1f ; movel %1@+,%0@+ ; movel %1@+,%0@+\n\t"
				"1:"
					: "=a" (d), "=a" (s), "=d" (count)
					: "0" (d), "1" (s), "2" (count)
			);
		} else {
		   long tmp;
		   __asm__ __volatile__(
					"movel  %0,%3\n\t"
					"lsrl   #1,%3 ; jcc 1f ; moveb %1@+,%0@+ ; subqw #1,%2\n\t"
					"lsrl   #1,%3 ; jcs 2f\n\t"  /* %0 increased=>bit 2 switched*/
					"movew  %1@+,%0@+  ; subqw  #2,%2 ; jra 2f\n\t"
				"1: lsrl   #1,%3 ; jcc 2f\n\t"
					"movew  %1@+,%0@+  ; subqw  #2,%2\n\t"
				"2: movew  %2,sp@-; lsrl #2,%2 ; jeq 6f\n\t"
					"lsrl   #1,%2 ; jcc 3f ; movel %1@+,%0@+\n\t"        
				"3: lsrl   #1,%2 ; jcc 4f ; movel %1@+,%0@+ ; movel %1@+,%0@+\n\t"
				"4: subql  #1,%2 ; jcs 6f\n\t"
				"5: movel  %1@+,%0@+;movel %1@+,%0@+\n\t"
					"movel  %1@+,%0@+;movel %1@+,%0@+\n\t"
					"dbra   %2,5b ; clrw %2; subql #1,%2; jcc 5b\n\t"
				"6: movew sp@+,%2; btst #1,%2 ; jeq 7f ; movew %1@+,%0@+\n\t"
				"7:              ; btst #0,%2 ; jeq 8f ; moveb %1@+,%0@+\n\t"
				"8:"
					: "=a" (d), "=a" (s), "=d" (count), "=d" (tmp)
					: "0" (d), "1" (s), "2" (count)
			);
		}
	} else {
		if (count < 16) {
			__asm__ __volatile__(
					"lsrl   #1,%2 ; jcc 1f ; moveb %1@-,%0@-\n\t"
				"1: lsrl   #1,%2 ; jcc 1f ; movew %1@-,%0@-\n\t"
				"1: lsrl   #1,%2 ; jcc 1f ; movel %1@-,%0@-\n\t"
				"1: lsrl   #1,%2 ; jcc 1f ; movel %1@-,%0@- ; movel %1@-,%0@-\n\t"
				"1:"
					: "=a" (d), "=a" (s), "=d" (count)
					: "0" ((char *) d + count), "1" ((char *) s + count), "2" (count)
			);
		} else {
			long tmp;
			__asm__ __volatile__(
					"movel %0,%3\n\t"
					"lsrl   #1,%3 ; jcc 1f ; moveb %1@-,%0@- ; subqw #1,%2\n\t"
					"lsrl   #1,%3 ; jcs 2f\n\t"  /* %0 increased=>bit 2 switched*/
					"movew  %1@-,%0@-  ; subqw  #2,%2 ; jra 2f\n\t"
				"1: lsrl   #1,%3 ; jcc 2f\n\t"
					"movew  %1@-,%0@-  ; subqw  #2,%2\n\t"
				"2: movew %2,sp@-; lsrl #2,%2 ; jeq 6f\n\t"
					"lsrl   #1,%2 ; jcc 3f ; movel %1@-,%0@-\n\t"        
				"3: lsrl   #1,%2 ; jcc 4f ; movel %1@-,%0@- ; movel %1@-,%0@-\n\t"
				"4: subql  #1,%2 ; jcs 6f\n\t"
				"5: movel %1@-,%0@-;movel %1@-,%0@-\n\t"
					"movel %1@-,%0@-;movel %1@-,%0@-\n\t"
					"dbra %2,5b ; clrw %2; subql #1,%2; jcc 5b\n\t"
				"6: movew sp@+,%2; btst #1,%2 ; jeq 7f ; movew %1@-,%0@-\n\t"
				"7:              ; btst #0,%2 ; jeq 8f ; moveb %1@-,%0@-\n\t"
				"8:"
					: "=a" (d), "=a" (s), "=d" (count), "=d" (tmp)
					: "0" ((char *) d + count), "1" ((char *) s + count), "2" (count)
			);
		}
	}

	return(0);
}


/* ====================================================================== */

/* amicon_XXX routines - interface used by the world
 *
 * This system is now divided into two levels because of complications
 * caused by hardware scrolling. Top level functions:
 *
 *		amicon_bmove(), amicon_clear(), amicon_putc()
 *
 * handles y values in range [0, scr_height-1] that correspond to real
 * screen positions. y_wrap shift means that first line of bitmap may be
 * anywhere on this display. These functions convert lineoffsets to
 * bitmap offsets and deal with the wrap-around case by splitting blits.
 *
 *		amicon_bmove_physical_8()	 -- These functions fast implementations
 *		amicon_clear_physical_8()	 -- of original amicon_XXX fns.
 *		amicon_putc_physical_8()	 -- (fontwidth != 8) may be added later
 *
 * WARNING:
 *
 * At the moment amicon_putc() cannot blit across vertical wrap boundary
 * Implies should only really hardware scroll in rows. Only reason for
 * restriction is simplicity & efficiency at the moment.
 */

static __inline__ int real_y(struct display *p, int y)
{
	int y_bottom_size = p->var.yoffset/p->fontheight;
	int y_top_size = (p->var.yres-p->var.yoffset)/p->fontheight;

	return((y < y_top_size) ? (y+y_bottom_size) : (y-y_top_size));
}


static int amicon_clear(struct condata *conp, int sy, int sx, int height,
								int width)
{
	int unit = conp-vc_cons;
	struct display *p = &disp[unit];

	if ((sy <= p->cursor_y) && (p->cursor_y < sy+height) &&
		 (sx <= p->cursor_x) && (p->cursor_x < sx+width))
		CURSOR_UNDRAWN();

	if (p->fontwidth == 8) {
		/* Split blits that cross physical y_wrap boundary */

		u_int y_break = (p->var.yres-p->var.yoffset)/p->fontheight;

		if ((sy < y_break) && (sy+height-1 >= y_break)) {
			u_int b = y_break-sy;
			p->dispsw->clear(conp, p, real_y(p, sy), sx, b, width);
			p->dispsw->clear(conp, p, real_y(p, sy+b), sx, height-b, width);
		} else
			p->dispsw->clear(conp, p, real_y(p, sy), sx, height, width);
	} else {
		/* No support (fontwidth != 8) yet */
	}

	return(0);
}


static int amicon_putc(struct condata *conp, int c, int y, int x)
{
	int unit = conp-vc_cons;
	struct display *p = &disp[unit];

	if ((p->cursor_x == x) && (p->cursor_y == y))
		CURSOR_UNDRAWN();
	
	if (p->fontwidth == 8)
		p->dispsw->putc(conp, p, c, real_y(p, y), x);
	else {
		/* Currently no support for (fontwidth != 8) */
	}

	return(0);
}


static int amicon_putcs(struct condata *conp, const char *s, int count, int y,
								int x)
{
	int unit = conp-vc_cons;
	struct display *p = &disp[unit];

	if ((p->cursor_y == y) && (x <= p->cursor_x) && (p->cursor_x < x+count))
		CURSOR_UNDRAWN();

	if (p->fontwidth == 8)
		p->dispsw->putcs(conp, p, s, count, real_y(p, y), x);
	else {
		/* Currently no support for (fontwidth != 8) */
	}

	return(0);
}


static int amicon_cursor(struct condata *conp, int mode)
{
	int unit = conp-vc_cons;
	struct display *p = &disp[unit];

	if (CURSOR_UNDRAWN ())
		if (p->fontwidth == 8)
			p->dispsw->rev_char(p, p->cursor_x, real_y(p, p->cursor_y));
		else {
			/* Currently no support for (fontwidth != 8) */
		}
	p->cursor_x = conp->vc_x;
	p->cursor_y = conp->vc_y;

	switch (mode) {
		case CM_ERASE:
			cursor_on = 0;
			break;

		case CM_MOVE:
		case CM_DRAW:
			vbl_cursor_cnt = CURSOR_DRAW_DELAY;
			cursor_on = 1;
			break;
	}

	return(0);

#if 0	/* ++Geert: Sorry, no hardware cursor at the moment */
	if (mode == CM_ERASE)
		p->cursor_visible = 0;
	else {
		p->cursor_visible = 1;
		p->cursor_latch = 0;
		p->cursor_x = (u_short)conp->vc_x;
		p->cursor_y = (u_short)conp->vc_y;
		p->cursor_latch = 1;
	}
#endif

	return(0);
}


static void amicon_vbl_handler(struct intframe *fp, void *dummy)
{
	struct display *p;

	if (!cursor_on)
		return;
	
	if (vbl_cursor_cnt && (--vbl_cursor_cnt == 0)) {
		/* Here no check is possible for console changing. The console
		 * switching code should set vbl_cursor_cnt to an appropriate value.
		 */
		p = &disp[fg_console];
		if (p->fontwidth == 8)
			p->dispsw->rev_char(p, p->cursor_x, real_y(p, p->cursor_y));
		else {
			/* Currently no support for (fontwidth != 8) */
		}
		cursor_drawn ^= 1;
		vbl_cursor_cnt = cursor_blink_rate;
	}
}


static int amicon_scroll(struct condata *conp, int t, int b, int dir,
								 int count)
{
	int unit = conp-vc_cons;
	struct display *p = &disp[unit];

	amicon_cursor(conp, CM_ERASE);

	/*
	 *	++Geert: Only use ywrap if the current screenmode supports it
	 *				AND the console is in text mode
	 */

	switch (dir) {
		case SM_UP:
			if ((t == 0) && (b == conp->vc_rows) && (p->ywrapstep == 1) &&
				 (vt_cons[unit].vc_mode == KD_TEXT)) {
				/* Can hardware ywrap */

				if (count > conp->vc_rows)				/* Maximum realistic size */
					count = conp->vc_rows;

				p->var.yoffset += p->fontheight*count;
				if (p->var.yoffset >= p->var.yres)	/* Deal with wrap */
					p->var.yoffset -= p->var.yres;

				p->var.vmode |= FB_VMODE_WRAP;
				fb_info->updatevar(unit);
			} else
				amicon_bmove(conp, t+count, 0, t, 0, b-t-count, conp->vc_cols);

			amicon_clear(conp, b-count, 0, count, conp->vc_cols);
			break;

		case SM_DOWN:
			if ((t == 0) && (b == conp->vc_rows) && (p->ywrapstep == 1) &&
				 (vt_cons[unit].vc_mode == KD_TEXT)) {
				/* Can hardware ywrap */

				if (count > conp->vc_rows)				/* Maximum realistic size */
					count = conp->vc_rows;

				p->var.yoffset -= p->fontheight*count;
				if (p->var.yoffset < 0)					/* Deal with wrap */
					p->var.yoffset += p->var.yres;

				p->var.vmode |= FB_VMODE_WRAP;
				fb_info->updatevar(unit);
			} else
				amicon_bmove(conp, t, 0, t+count, 0, b-t-count, conp->vc_cols);

			/* Fixed bmove() should end Arno's frustration with copying?
			 * Confusius say:
			 *		Man who copies in wrong direction, end up with trashed data
			 */
			amicon_clear(conp, t, 0, count, conp->vc_cols);
			break;

		case SM_LEFT:
			amicon_bmove(conp, 0, t+count, 0, t, conp->vc_rows, b-t-count);
			amicon_clear(conp, 0, b-count, conp->vc_rows, count);
			break;

		case SM_RIGHT:
			amicon_bmove(conp, 0, t, 0, t+count, conp->vc_rows, b-t-count);
			amicon_clear(conp, 0, t, conp->vc_rows, count);
			break;
	}

	return(0);
}


static int amicon_bmove(struct condata *conp, int sy, int sx, int dy, int dx,
								int height, int width)
{
	int unit = conp-vc_cons;
	struct display *p = &disp[unit];
	u_int y_break;

	if (((sy <= p->cursor_y) && (p->cursor_y < sy+height) &&
		  (sx <= p->cursor_x) && (p->cursor_x < sx+width)) ||
		 ((dy <= p->cursor_y) && (p->cursor_y < dy+height) &&
		  (dx <= p->cursor_x) && (p->cursor_x < dx+width)))
		amicon_cursor(conp, CM_ERASE);

	if (p->fontwidth == 8) {
		/* Split blits that cross physical y_wrap case.
		 * Pathological case involves 4 blits, better to use recursive
		 * code rather than unrolled case
		 *
		 * Recursive invocations don't need to erase the cursor over and
		 * over again, so we use amicon_bmove_rec()
		 */
		y_break = (p->var.yres-p->var.yoffset)/p->fontheight;
		amicon_bmove_rec(p, sy, sx, dy, dx, height, width, y_break);
	} else {
		/* No support (fontwidth != 8) yet */
	}

	return(0);
}


static void amicon_bmove_rec(struct display *p, int sy, int sx, int dy, int dx,
									  int height, int width, u_int y_break)
{
	u_int b;

	if ((sy < y_break) && (sy+height > y_break)) {
		b = y_break-sy;
		if (dy < sy) {			/* Avoid trashing self */
			amicon_bmove_rec(p, sy, sx, dy, dx, b, width, y_break);
			amicon_bmove_rec(p, sy+b, sx, dy+b, dx, height-b, width, y_break);
		} else {
			amicon_bmove_rec(p, sy+b, sx, dy+b, dx, height-b, width, y_break);
			amicon_bmove_rec(p, sy, sx, dy, dx, b, width, y_break);
		}
		return;
	}

	if ((dy < y_break) && (dy+height > y_break)) {
		b = y_break-dy;
		if (dy < sy) {			/* Avoid trashing self */
			amicon_bmove_rec(p, sy, sx, dy, dx, b, width, y_break);
			amicon_bmove_rec(p, sy+b, sx, dy+b, dx, height-b, width, y_break);
		} else {
			amicon_bmove_rec(p, sy+b, sx, dy+b, dx, height-b, width, y_break);
			amicon_bmove_rec(p, sy, sx, dy, dx, b, width, y_break);
		}
		return;
	}
	p->dispsw->bmove(p, real_y(p, sy), sx, real_y(p, dy), dx, height, width);
}


static int amicon_switch(struct condata *conp)
{
	if (!(conp->vc_flags & CON_INITED))
		return(-1);

	if (fb_info && fb_info->switch_con)
		(*fb_info->switch_con)(conp-vc_cons);
	return(0);
}


static int amicon_blank(int blank)
{
	(*fb_info->blank)(blank);
	return(0);
}


/* ====================================================================== */

/*
 *		Low Level Operations for the various display memory organizations.
 *
 *		Currently only the following organizations are supported here:
 *
 *		  - Monochrome
 *		  - Color Interleaved Planes  la Amiga
 *		  - Color Normal Planes
 *
 *		The other organizations are still in atari/atacon.c.
 *
 *		This means this code is (almost) ready for the big atacon/amicon merge
 *		operation.
 *
 *		P.S. What about graphics boards with a `real' textmode?
 *
 *																						Geert
 */


	/*
	 *		Support for Monochrome
	 *
	 *		In fact this is the `shared' case for both Interleaved Planes and
	 *		Normal Planes if bits_per_pixel == 1.
	 */

static void amicon_bmove_mono(struct display *p, int sy, int sx, int dy, int dx,
										int height, int width)
{
	u_char *src, *dest;
	u_int rows;

	if ((sx == 0) && (sy == 0) && (width == p->next_line)) {
		src = p->screen_base;
		dest = p->screen_base+dy*p->fontheight*width;
		mymemmove(dest, src, height*p->fontheight*width);
	} else if (dy <= sy) {
		src = p->screen_base+sy*p->fontheight*p->next_line+sx;
		dest = p->screen_base+dy*p->fontheight*p->next_line+dx;
		for (rows = height*p->fontheight; rows--;) {
			mymemmove(dest, src, width);
			src += p->next_line;
			dest += p->next_line;
		}
	} else {
		src = p->screen_base+((sy+height)*p->fontheight-1)*p->next_line+sx;
		dest = p->screen_base+((dy+height)*p->fontheight-1)*p->next_line+dx;
		for (rows = height*p->fontheight; rows--;) {
			mymemmove(dest, src, width);
			src -= p->next_line;
			dest -= p->next_line;
		}
	}
}


static void amicon_clear_mono(struct condata *conp, struct display *p, int sy,
										int sx, int height, int width)
{
	u_char *dest;
	u_int rows;

	dest = p->screen_base+sy*p->fontheight*p->next_line+sx;

	if ((sx == 0) && (width == p->next_line))
		if (attr_reverse(conp->vc_attr))
			mymemset(dest, height*p->fontheight*width);
		else
			mymemclear(dest, height*p->fontheight*width);
	else
		for (rows = height*p->fontheight; rows--; dest += p->next_line)
			if (attr_reverse(conp->vc_attr))
				mymemset(dest, width);
			else
				mymemclear_small(dest, width);
}


static void amicon_putc_mono(struct condata *conp, struct display *p, int c,
									  int y, int x)
{
	u_char *dest, *cdat;
	u_int rows, bold, reverse, underline;
	u_char d;

	if ((c >= 32) && (c < 127))
		/* lower half of charset -- normal ASCII */
		c -= 32;
	else if ((c >= 160) && (c < 255))
		/* upper half of charset -- special characters */
		c -= 64;
	else
		/* not defined in charset -- non-printable character */
		return;

	dest = p->screen_base+y*p->fontheight*p->next_line+x;
	cdat = p->fontdata+(c*p->fontheight);
	bold = attr_bold(conp->vc_attr);
	reverse = attr_reverse(conp->vc_attr);
	underline = attr_underline(conp->vc_attr);

	for (rows = p->fontheight; rows--; dest += p->next_line) {
  		d = *cdat++;
		if (underline && !rows)
			d = 0xff;
		else if (bold)
			d |= d>>1;
		if (reverse)
			d = ~d;
		*dest = d;
	}
}


static void amicon_putcs_mono(struct condata *conp, struct display *p,
										const char *s, int count, int y, int x)
{
	u_char *dest, *dest0, *cdat;
	u_int rows, bold, reverse, underline;
	u_char c, d;

	dest0 = p->screen_base+y*p->fontheight*p->next_line+x;
	bold = attr_bold(conp->vc_attr);
	reverse = attr_reverse(conp->vc_attr);
	underline = attr_underline(conp->vc_attr);

	while (count--) {
		c = *s++;
		dest = dest0++;
		if ((c >= 32) && (c < 127))
			/* lower half of charset -- normal ASCII */
			c -= 32;
		else if ((c >= 160) && (c < 255))
			/* upper half of charset -- special characters */
			c -= 64;
		else
			continue;
		cdat = p->fontdata+(c*p->fontheight);
		for (rows = p->fontheight; rows--; dest += p->next_line) {
 			d = *cdat++;
			if (underline && !rows)
				d = 0xff;
			else if (bold)
				d |= d>>1;
			if (reverse)
				d = ~d;
			*dest = d;
		}
	}
}


static void amicon_rev_char_mono(struct display *p, int x, int y)
{
	u_char *dest;
	u_int rows;

	dest = p->screen_base+y*p->fontheight*p->next_line+x;
	for (rows = p->fontheight; rows--; dest += p->next_line)
		*dest = ~*dest;
}


/* ====================================================================== */

	/*
	 *		Support for Color Interleaved Planes
	 *
	 *		This code heavily relies on the fact that
	 *
	 *			next_line == interleave == next_plane*bits_per_pixel
	 *
	 *		But maybe it can be merged with the code for normal bitplanes without
	 *		much performance loss?
	 */

static void amicon_bmove_ilbm(struct display *p, int sy, int sx, int dy, int dx,
										int height, int width)
{
	if ((sx == 0) && (sy == 0) && (width == p->next_plane))
		mymemmove(p->screen_base+dy*p->fontheight*p->next_line, p->screen_base,
					 height*p->fontheight*p->next_line);
	else {
		u_char *src, *dest;
		u_int i;
		
		if (dy <= sy) {
			src = p->screen_base+sy*p->fontheight*p->next_line+sx;
			dest = p->screen_base+dy*p->fontheight*p->next_line+dx;
			for (i = p->var.bits_per_pixel*height*p->fontheight; i--;) {
				mymemmove(dest, src, width);
				src += p->next_plane;
				dest += p->next_plane;
			}
		} else {
			src = p->screen_base+(sy+height)*p->fontheight*p->next_line+sx;
			dest = p->screen_base+(dy+height)*p->fontheight*p->next_line+dx;
			for (i = p->var.bits_per_pixel*height*p->fontheight; i--;) {
				src -= p->next_plane;
				dest -= p->next_plane;
				mymemmove(dest, src, width);
			}
		}
	}
}


static void amicon_clear_ilbm(struct condata *conp, struct display *p, int sy,
										int sx, int height, int width)
{
	u_char *dest;
	u_int i, rows;
	int bg, bg0;

	dest = p->screen_base+sy*p->fontheight*p->next_line+sx;

	bg0 = (conp->vc_video_erase_char>>12) & 0x0f;
	for (rows = height*p->fontheight; rows--;) {
		bg = bg0;
		for (i = p->var.bits_per_pixel; i--; dest += p->next_plane) {
			if (bg & 1)
				mymemset(dest, width);
			else
				mymemclear(dest, width);
			bg >>= 1;
		}
	}
}


static void amicon_putc_ilbm(struct condata *conp, struct display *p, int c,
									  int y, int x)
{
	u_char *dest, *cdat;
	u_int rows, i;
	u_char d;
	int fg0, bg0, fg, bg;

	if ((c >= 32) && (c < 127))
		/* lower half of charset -- normal ASCII */
		c -= 32;
	else if ((c >= 160) && (c < 255))
		/* upper half of charset -- special characters */
		c -= 64;
	else
		/* not defined in charset -- non-printable character */
		return;

	dest = p->screen_base+y*p->fontheight*p->next_line+x;
	cdat = p->fontdata+(c*p->fontheight);
	fg0 = attr_fgcol(conp->vc_attr);
	bg0 = attr_bgcol(conp->vc_attr);

	for (rows = p->fontheight; rows--;) {
		d = *cdat++;
		fg = fg0;
		bg = bg0;
		for (i = p->var.bits_per_pixel; i--; dest += p->next_plane) {
			if (bg & 1)
				if (fg & 1)
					*dest = 0xff;
				else
					*dest = ~d;
			else
				if (fg & 1)
					*dest = d;
				else
					*dest = 0x00;
			bg >>= 1;
			fg >>= 1;
		}
	}
}


/*
 *		I splitted the console character loop in two parts:
 *
 *		  - slow version: this blits one character at a time
 *
 *		  - fast version: this blits 4 characters at a time at a longword aligned
 *								address, to reduce the number of expensive Chip RAM
 *								accesses.
 *
 *		Experiments on my A4000/040 revealed that this makes a console switch on a
 *		640x400 screen with 256 colors about 3 times faster.
 *
 *																						Geert
 */

static void amicon_putcs_ilbm(struct condata *conp, struct display *p,
										const char *s, int count, int y, int x)
{
	u_char *dest0, *dest, *cdat1, *cdat2, *cdat3, *cdat4;
	u_int rows, i;
	u_char c1, c2, c3, c4;
	u_long d;
	int fg0, bg0, fg, bg;

	dest0 = p->screen_base+y*p->fontheight*p->next_line+x;
	fg0 = attr_fgcol(conp->vc_attr);
	bg0 = attr_bgcol(conp->vc_attr);

	while (count--)
		if ((x&3) || (count < 3)) {	/* Slow version */
slow_version:
			c1 = *s++;
			dest = dest0++;
			x++;

			if ((c1 >= 32) && (c1 < 127))
				/* lower half of charset -- normal ASCII */
				c1 -= 32;
			else if ((c1 >= 160) && (c1 < 255))
				/* upper half of charset -- special characters */
				c1 -= 64;
			else
				/* not defined in charset -- non-printable character */
				continue;

			cdat1 = p->fontdata+(c1*p->fontheight);
			for (rows = p->fontheight; rows--;) {
				d = *cdat1++;
				fg = fg0;
				bg = bg0;
				for (i = p->var.bits_per_pixel; i--; dest += p->next_plane) {
					if (bg & 1)
						if (fg & 1)
							*dest = 0xff;
						else
							*dest = ~d;
					else
						if (fg & 1)
							*dest = d;
						else
							*dest = 0x00;
					bg >>= 1;
					fg >>= 1;
				}
			}
		} else {								/* Fast version */
			c1 = s[0];
			c2 = s[1];
			c3 = s[2];
			c4 = s[3];

			if ((c1 >= 32) && (c1 < 127))
				c1 -= 32;
			else if ((c1 >= 160) && (c1 < 255))
				c1 -= 64;
			else {
				s++;
				dest0++;
				x++;
				continue;
			}
			if ((c2 >= 32) && (c2 < 127))
				c2 -= 32;
			else if ((c2 >= 160) && (c2 < 255))
				c2 -= 64;
			else
				goto slow_version;
			if ((c3 >= 32) && (c3 < 127))
				c3 -= 32;
			else if ((c3 >= 160) && (c3 < 255))
				c3 -= 64;
			else
				goto slow_version;
			if ((c4 >= 32) && (c4 < 127))
				c4 -= 32;
			else if ((c4 >= 160) && (c4 < 255))
				c4 -= 64;
			else
				goto slow_version;

			dest = dest0;
			cdat1 = p->fontdata+(c1*p->fontheight);
			cdat2 = p->fontdata+(c2*p->fontheight);
			cdat3 = p->fontdata+(c3*p->fontheight);
			cdat4 = p->fontdata+(c4*p->fontheight);
			for (rows = p->fontheight; rows--;) {
				d = *cdat1++<<24 | *cdat2++<<16 | *cdat3++<<8 | *cdat4++;
				fg = fg0;
				bg = bg0;
				for (i = p->var.bits_per_pixel; i--; dest += p->next_plane) {
					if (bg & 1)
						if (fg & 1)
							*(u_long *)dest = 0xffffffff;
						else
							*(u_long *)dest = ~d;
					else
						if (fg & 1)
							*(u_long *)dest = d;
						else
							*(u_long *)dest = 0x00000000;
					bg >>= 1;
					fg >>= 1;
				}
			}
			s += 4;
			dest0 += 4;
			x += 4;
			count -= 3;
		}
}


static void amicon_rev_char_ilbm(struct display *p, int x, int y)
{
	u_char *dest, *dest0;
	u_int rows, i;
	int mask;

	dest0 = p->screen_base+y*p->fontheight*p->next_line+x;
	mask = p->fgcol ^ p->bgcol;

	/*
	 *		This should really obey the individual character's
	 *		background and foreground colors instead of simply
	 *		inverting.
	 */

	for (i = p->var.bits_per_pixel; i--; dest0 += p->next_plane) {
		if (mask & 1) {
			dest = dest0;
			for (rows = p->fontheight; rows--; dest += p->next_line)
				*dest = ~*dest;
		}
		mask >>= 1;
	}
}


/* ====================================================================== */

	/*
	 *		Support for Color Planes
	 */

static void amicon_bmove_plan(struct display *p, int sy, int sx, int dy, int dx,
										int height, int width)
{
	u_char *src, *dest, *src0, *dest0;
	u_int i, rows;

	if ((sx == 0) && (sy == 0) && (width == p->next_line)) {
		src = p->screen_base;
		dest = p->screen_base+dy*p->fontheight*width;
		for (i = p->var.bits_per_pixel; i--;) {
			mymemmove(dest, src, height*p->fontheight*width);
			src += p->next_plane;
			dest += p->next_plane;
		}
	} else if (dy <= sy) {
		src0 = p->screen_base+sy*p->fontheight*p->next_line+sx;
		dest0 = p->screen_base+dy*p->fontheight*p->next_line+dx;
		for (i = p->var.bits_per_pixel; i--;) {
			src = src0;
			dest = dest0;
			for (rows = height*p->fontheight; rows--;) {
				mymemmove(dest, src, width);
				src += p->next_line;
				dest += p->next_line;
			}
			src0 += p->next_plane;
			dest0 += p->next_plane;
		}
	} else {
		src0 = p->screen_base+(sy+height)*p->fontheight*p->next_line+sx;
		dest0 = p->screen_base+(dy+height)*p->fontheight*p->next_line+dx;
		for (i = p->var.bits_per_pixel; i--;) {
			src = src0;
			dest = dest0;
			for (rows = height*p->fontheight; rows--;) {
				src -= p->next_line;
				dest -= p->next_line;
				mymemmove(dest, src, width);
			}
			src0 += p->next_plane;
			dest0 += p->next_plane;
		}
	}
}


static void amicon_clear_plan(struct condata *conp, struct display *p, int sy,
										int sx, int height, int width)
{
	u_char *dest, *dest0;
	u_int i, rows;
	int bg;

	dest0 = p->screen_base+sy*p->fontheight*p->next_line+sx;

	bg = (conp->vc_video_erase_char>>12) & 0x0f;
	for (i = p->var.bits_per_pixel; i--; dest0 += p->next_plane) {
		dest = dest0;
		for (rows = height*p->fontheight; rows--; dest += p->next_line)
			if (bg & 1)
				mymemset(dest, width);
			else
				mymemclear(dest, width);
		bg >>= 1;
	}
}


static void amicon_putc_plan(struct condata *conp, struct display *p, int c,
									  int y, int x)
{
	u_char *dest, *dest0, *cdat, *cdat0;
	u_int rows, i;
	u_char d;
	int fg, bg;

	if ((c >= 32) && (c < 127))
		/* lower half of charset -- normal ASCII */
		c -= 32;
	else if ((c >= 160) && (c < 255))
		/* upper half of charset -- special characters */
		c -= 64;
	else
		/* not defined in charset -- non-printable character */
		return;

	dest0 = p->screen_base+y*p->fontheight*p->next_line+x;
	cdat0 = p->fontdata+(c*p->fontheight);
	fg = attr_fgcol(conp->vc_attr);
	bg = attr_bgcol(conp->vc_attr);

	for (i = p->var.bits_per_pixel; i--; dest0 += p->next_plane) {
		dest = dest0;
		cdat = cdat0;
		for (rows = p->fontheight; rows--; dest += p->next_line) {
			d = *cdat++;
			if (bg & 1)
				if (fg & 1)
					*dest = 0xff;
				else
					*dest = ~d;
			else
				if (fg & 1)
					*dest = d;
				else
					*dest = 0x00;
		}
		bg >>= 1;
		fg >>= 1;
	}
}


/*
 *		I splitted the console character loop in two parts
 *		(cfr. amicon_putcs_ilbm())
 */

static void amicon_putcs_plan(struct condata *conp, struct display *p,
										const char *s, int count, int y, int x)
{
	u_char *dest, *dest0, *dest1;
	u_char *cdat1, *cdat2, *cdat3, *cdat4, *cdat10, *cdat20, *cdat30, *cdat40;
	u_int rows, i;
	u_char c1, c2, c3, c4;
	u_long d;
	int fg0, bg0, fg, bg;

	dest0 = p->screen_base+y*p->fontheight*p->next_line+x;
	fg0 = attr_fgcol(conp->vc_attr);
	bg0 = attr_bgcol(conp->vc_attr);

	while (count--)
		if ((x&3) || (count < 3)) {	/* Slow version */
slow_version:
			c1 = *s++;
			dest1 = dest0++;
			x++;

			if ((c1 >= 32) && (c1 < 127))
				/* lower half of charset -- normal ASCII */
				c1 -= 32;
			else if ((c1 >= 160) && (c1 < 255))
				/* upper half of charset -- special characters */
				c1 -= 64;
			else
				/* not defined in charset -- non-printable character */
				continue;

			cdat10 = p->fontdata+(c1*p->fontheight);
			fg = fg0;
			bg = bg0;

			for (i = p->var.bits_per_pixel; i--; dest1 += p->next_plane) {
				dest = dest1;
				cdat1 = cdat10;
				for (rows = p->fontheight; rows--; dest += p->next_line) {
					d = *cdat1++;
					if (bg & 1)
						if (fg & 1)
							*dest = 0xff;
						else
							*dest = ~d;
					else
						if (fg & 1)
							*dest = d;
						else
							*dest = 0x00;
				}
				bg >>= 1;
				fg >>= 1;
			}
		} else {								/* Fast version */
			c1 = s[0];
			c2 = s[1];
			c3 = s[2];
			c4 = s[3];

			if ((c1 >= 32) && (c1 < 127))
				c1 -= 32;
			else if ((c1 >= 160) && (c1 < 255))
				c1 -= 64;
			else {
				s++;
				dest0++;
				x++;
				continue;
			}
			if ((c2 >= 32) && (c2 < 127))
				c2 -= 32;
			else if ((c2 >= 160) && (c2 < 255))
				c2 -= 64;
			else
				goto slow_version;
			if ((c3 >= 32) && (c3 < 127))
				c3 -= 32;
			else if ((c3 >= 160) && (c3 < 255))
				c3 -= 64;
			else
				goto slow_version;
			if ((c4 >= 32) && (c4 < 127))
				c4 -= 32;
			else if ((c4 >= 160) && (c4 < 255))
				c4 -= 64;
			else
				goto slow_version;

			dest1 = dest0;
			cdat10 = p->fontdata+(c1*p->fontheight);
			cdat20 = p->fontdata+(c2*p->fontheight);
			cdat30 = p->fontdata+(c3*p->fontheight);
			cdat40 = p->fontdata+(c4*p->fontheight);
			fg = fg0;
			bg = bg0;

			for (i = p->var.bits_per_pixel; i--; dest1 += p->next_plane) {
				dest = dest1;
				cdat1 = cdat10;
				cdat2 = cdat20;
				cdat3 = cdat30;
				cdat4 = cdat40;
				for (rows = p->fontheight; rows--; dest += p->next_line) {
					d = *cdat1++<<24 | *cdat2++<<16 | *cdat3++<<8 | *cdat4++;
					if (bg & 1)
						if (fg & 1)
							*(u_long *)dest = 0xffffffff;
						else
							*(u_long *)dest = ~d;
					else
						if (fg & 1)
							*(u_long *)dest = d;
						else
							*(u_long *)dest = 0x00000000;
				}
				bg >>= 1;
				fg >>= 1;
			}
			s += 4;
			dest0 += 4;
			x += 4;
			count -= 3;
		}
}


static void amicon_rev_char_plan(struct display *p, int x, int y)
{
	u_char *dest, *dest0;
	u_int rows, i;
	int mask;

	dest0 = p->screen_base+y*p->fontheight*p->next_line+x;
	mask = p->fgcol ^ p->bgcol;

	/*
	 *		This should really obey the individual character's
	 *		background and foreground colors instead of simply
	 *		inverting.
	 */

	for (i = p->var.bits_per_pixel; i--; dest0 += p->next_plane) {
		if (mask & 1) {
			dest = dest0;
			for (rows = p->fontheight; rows--; dest += p->next_line)
				*dest = ~*dest;
		}
		mask >>= 1;
	}
}


/* ====================================================================== */

	/*
	 *		The console `switch' structure for the Amiga native console
	 */

struct consw ami_con =
{
	amicon_init, amicon_deinit, amicon_clear, amicon_putc, amicon_putcs,
	amicon_cursor, amicon_scroll, amicon_bmove, amicon_switch, amicon_blank
};
