/*
 * atari/atafb.c -- Low level implementation of Atari frame buffer device
 * amiga/amifb.c -- Low level implementation of Amiga frame buffer device
 *
 *  Copyright (C) 1994 Martin Schaller & Roman Hodek & Geert Uytterhoeven
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 * History:
 *	  - 03 Jan 95: Original version my Martin Schaller: The TT driver and
 *						all the device independant stuff
 *	  - 09 Jan 95: Roman: I've added the hardware abstraction (hw_switch)
 *						and wrote the Falcon, ST(E), and External drivers
 *						based on the original TT driver.
 *	  - 26 Jan 95: Geert: Amiga version
 *	  - 19 Feb 95: Hamish: added Jes Sorensen's ECS patches to the Amiga
 *						frame buffer device.  This provides ECS support and the
 *						following screen-modes: multiscan, multiscan-lace,
 *						super72, super72-lace, dblntsc, dblpal & euro72.
 *						He suggests that we remove the old AGA screenmodes,
 *						as they are non-standard, and some of them don't work
 *						under ECS.
 *	  - 24 Jul 95: Geert: complete rewrite to support color, virtual screens,
 *						dynamic mode configuration, \ldots (AGA version)
 */




/* --- BEGIN: TO DO --- */
/*

	- OCS and ECS (change gunbits to 4 !!!)

	- hardware cursor

	- PAL/NTSC encodings (a.o. BMC0_PAL)

	- Interlaced screen -> Interlaced sprite/hardware cursor

	- check internal error return codes

*/
/* --- END: TO DO --- */


#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/amigahw.h>
#include <linux/bootinfo.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/config.h>
#include <linux/interrupt.h>
#include <linux/amigaints.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "amifb.h"




/*******************************************************************************


	Generic video timings
	---------------------

	Timings used by the frame buffer interface:

	+----------+---------------------------------------------+----------+-------+
	|          |                ^                            |          |       |
	|          |                |upper_margin                |          |       |
	|          |                                            |          |       |
	+----------+---------------------------------------------+----------+-------+
	|          |                ^                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|   left   |                |                            |  right   | hsync |
	|  margin  |                |       xres                 |  margin  |  len  |
	|<-------->|<---------------+--------------------------->|<-------->|<----->|
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |yres                        |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                |                            |          |       |
	|          |                                            |          |       |
	+----------+---------------------------------------------+----------+-------+
	|          |                ^                            |          |       |
	|          |                |lower_margin                |          |       |
	|          |                                            |          |       |
	+----------+---------------------------------------------+----------+-------+
	|          |                ^                            |          |       |
	|          |                |vsync_len                   |          |       |
	|          |                                            |          |       |
	+----------+---------------------------------------------+----------+-------+


	Amiga video timings
	-------------------

	The Amiga native chipsets uses another timing scheme:

		- hsstrt:	Start of horizontal synchronization pulse
		- hsstop:	End of horizontal synchronization pulse
		- htotal:	Last value on the line (i.e. line length = htotal+1)
		- vsstrt:	Start of vertical synchronization pulse
		- vsstop:	Start of vertical synchronization pulse
		- vtotal:	Last line value (i.e. number of lines = vtotal+1)
		- hcenter:	Start of vertical retrace for interlace??

	You can specify the blanking timings independently. Currently I just set
	them equal to the respective synchronization values:

		- hbstrt:	Start of horizontal blank
		- hbstop:	End of horizontal blank
		- vbstrt:	Start of vertical blank
		- vbstop:	Start of vertical blank

	Horizontal values are in color clock cycles (280 ns), vertical values are in
	scanlines.

	(0, 0) is somewhere in the upper-left corner :-)


	Amiga visible window definitions
	--------------------------------

	Currently I only have values for AGA, SHRES (28 MHz dotclock). Feel free to
	make corrections and/or additions.

	Within the above synchronization specifications, the visible window is
	defined by the following parameters (actual register resolutions may be
	different; all horizontal values are normalized to the pixel clock):

		- diwstrt_h:	Horizontal start of the visible window
		- diwstop_h:	Horizontal stop+1(*) of the visible window
		- diwstrt_v:	Vertical start of the visible window
		- diwstop_v:	Vertical stop of the visible window
		- ddfstrt:		Horizontal start of display DMA
		- ddfstop:		Horizontal stop of display DMA
		- hscroll:		Horizontal display output delay

	Sprite positioning:

		- sprstrt_h:	Horizontal start-4 of sprite
		- sprstrt_v:	Vertical start of sprite

	(*) Even Commodore did it wrong in the AGA monitor drivers by not adding 1.

	Horizontal values are in dotclock cycles (35 ns), vertical values are in
	scanlines.

	(0, 0) is somewhere in the upper-left corner :-)


	Dependencies (AGA, SHRES (35 ns dotclock))
	-------------------------------------------

	Since there are much more parameters for the Amiga display than for the
	frame buffer interface, there must be some depencies among the Amiga display
	parameters. Here's what I found out:

		- ddfstrt and ddfstop are best aligned to 64 pixels.
		- the chipset needs 64+4 horizontal pixels after the DMA start before the
		  first pixel is output, so diwstrt_h = ddfstrt+64+4 if you want to
		  display the first pixel on the line too. Increase diwstrt_h for virtual
		  screen panning.
		- the display DMA always fetches 64 pixels at a time (fmode = 3).
		- ddfstop is ddfstrt+#pixels-64.
		- diwstop_h = diwstrt_h+xres+1. Because of the additional 1 this can be 1
		  more than htotal.
		- hscroll simply adds a delay to the display output. Smooth horizontal
		  panning needs an extra 64 pixels on the left to prefetch the pixels that
		  `fall off' on the left.
		- if ddfstrt < 192, the sprite DMA cycles are all stolen by the bitplane
		  DMA, so it's best to make the DMA start as late as possible.
		- you really don't want to make ddfstrt < 128.
		- I make diwstop_h and diwstop_v as large as possible.


	DMA priorities
	--------------

	Since there are limits on the earliest start value for display DMA and the
	display of sprites, I use the following policy on horizontal panning and
	the hardware cursor:

		- if you want to start display DMA too early, you loose the ability to
		  do smooth horizontal panning (xpanstep 1 -> 64).
		- if you want to go even further, you loose the hardware cursor too.

	IMHO a hardware cursor is more important for X than horizontal scrolling,
	so that's my motivation.


	Implementation
	--------------

	AGA_decode_var() converts the frame buffer values to the Amiga values. It's
	just a `straightforward' implementation of the above rules.


	Standard VGA timings
	--------------------

					xres  yres    left  right  upper  lower    hsync    vsync
					----  ----    ----  -----  -----  -----    -----    -----
		80x25		 720   400      27     45     35     12      108        2
		80x30		 720   480      27     45     30      9      108        2

	These were taken from a XFree86 configuration file, recalculated for a 28 MHz
	dotclock (Amigas don't have a 25 MHz dotclock) and converted to frame buffer
	generic timings.

	As a comparison, graphics/monitor.h suggests the following:

					xres  yres    left  right  upper  lower    hsync    vsync
					----  ----    ----  -----  -----  -----    -----    -----

		VGA		 640   480      52    112     24     19    112 -      2 +
		VGA70		 640   400      52    112     27     21    112 -      2 -


	Sync polarities
	---------------

		VSYNC    HSYNC    Vertical size    Vertical total
		-----    -----    -------------    --------------
		  +        +           Reserved          Reserved
	     +        -                400               414
	     -        +                350               362
	     -        -                480               496

	Source: CL-GD542X Technical Reference Manual, Cirrus Logic, Oct 1992


	Broadcast video timings (not yet implemented)
	-----------------------

	Since broadcast video timings are `fixed' and only depend on the video
	system (PAL/NTSC), hsync_len and vsync_len are not used and must be set to
	zero. All xres/yres and margin values are defined within the `visible
	rectangle' of the display.

	According to the CCIR and RETMA specifications, we have the following values:

	CCIR -> PAL
	-----------

		- a scanline is 64 s long, of which 52.48 s are visible.
		- we have 625 scanlines, of which 575 are visible (interlaced); after
		  rounding this becomes 576.

	RETMA -> NTSC
	-------------

		- a scanline is XX s long, of which XX s are visible (please fill in!).
		- we have 525 scanlines, of which 484(?) are visible (interlaced).

	Thus if you want a PAL compatible display, you have to do the following:

		- set the FB_SYNC_BROADCAST flag to indicate that standard broadcast
		  timings are to be used.
		- make sure upper_margin+yres+lower_margin = 576 for an interlaced, and
		  288 for a non-interlaced display.
		- make sure (left_margin+xres+right_margin)*pixclock is a reasonable
		  approximation to 52.48 s.

	The settings for a NTSC compatible display are straightforward.

	Note that in a strict sense the PAL and NTSC standards only define the
	encoding of the color part (chrominance) of the video signal and don't say
	anything about horizontal/vertical synchronization nor refresh rates.


																				-- Geert --

*******************************************************************************/


	/*
	 *		Custom Chipset Definitions
	 */

#define CUSTOM_OFS(fld) ((long)&((struct CUSTOM*)0)->fld)


	/*
	 *		BPLCON0 -- Bitplane Control Register 0
	 */

#define	BPC0_HIRES		0x8000
#define	BPC0_BPU3		0x0010	/* AGA */
#define	BPC0_BPU2		0x4000	/* Bit plane used count */
#define	BPC0_BPU1		0x2000
#define	BPC0_BPU0		0x1000
#define	BPC0_HAM			0x0800	/* HAM mode */
#define	BPC0_DPF			0x0400	/* Double playfield */
#define	BPC0_COLOR		0x0200	/* Enable colorburst */
#define	BPC0_GAUD		0x0100	/* Genlock audio enable */
#define	BPC0_UHRES		0x0080	/* Ultrahi res enable */
#define	BPC0_SHRES		0x0040	/* Super hi res mode */
#define	BPC0_BYPASS		0x0020	/* Bypass LUT - AGA */
#define	BPC0_LPEN		0x0008	/* Light pen enable */
#define	BPC0_LACE		0x0004	/* Interlace */
#define	BPC0_ERSY		0x0002	/* External resync */
#define	BPC0_ECSENA		0x0001	/* ECS emulation disable */


	/*
	 *		BPLCON2 -- Bitplane Control Register 2
	 */

#define	BPC2_KILLEHB	0x0200	/* Kill EHB mode - AGA */
#define	BPC2_RDRAM		0x0100	/* Color table accesses read, not write - AGA */
#define	BPC2_SOGEN		0x0080	/* SOG output pin high - AGA */
#define	BPC2_PF2PRI		0x0040	/* PF2 priority over PF1 */
#define	BPC2_PF2P2		0x0020	/* PF2 priority wrt sprites */
#define	BPC2_PF2P1		0x0010
#define	BPC2_PF2P0		0x0008
#define	BPC2_PF1P2		0x0004	/* ditto PF1 */
#define	BPC2_PF1P1		0x0002
#define	BPC2_PF1P0		0x0001


	/*
	 *		BPLCON3 -- Bitplane Control Register 3 (AGA)
	 */

#define	BPC3_BANK2		0x8000	/* Bits to select color register bank */
#define	BPC3_BANK1		0x4000
#define	BPC3_BANK0		0x2000
#define	BPC3_PF2OF2		0x1000	/* Bits for color table offset when PF2 */
#define	BPC3_PF2OF1		0x0800
#define	BPC3_PF2OF0		0x0400
#define	BPC3_LOCT		0x0200	/* Color register writes go to low bits */
#define	BPC3_SPRES1		0x0080	/* Sprite resolution bits */
#define	BPC3_SPRES0		0x0040
#define	BPC3_BRDRBLNK	0x0020	/* Border blanked? */
#define	BPC3_BRDRTRAN	0x0010	/* Border transparent? */
#define	BPC3_BRDRSPRT	0x0002	/* Sprites in border? */
#define	BPC3_EXTBLKEN	0x0001	/* BLANK programmable */


	/*
	 *		BPLCON4 -- Bitplane Control Register 4 (AGA)
	 */

#define	BPC4_BPLAM7		0x8000	/* bitplane color XOR field */
#define	BPC4_BPLAM6		0x4000
#define	BPC4_BPLAM5		0x2000
#define	BPC4_BPLAM4		0x1000
#define	BPC4_BPLAM3		0x0800
#define	BPC4_BPLAM2		0x0400
#define	BPC4_BPLAM1		0x0200
#define	BPC4_BPLAM0		0x0100
#define	BPC4_ESPRM7		0x0080	/* 4 high bits for even sprite colors */
#define	BPC4_ESPRM6		0x0040
#define	BPC4_ESPRM5		0x0020
#define	BPC4_ESPRM4		0x0010
#define	BPC4_OSPRM7		0x0008	/* 4 high bits for odd sprite colors */
#define	BPC4_OSPRM6		0x0004
#define	BPC4_OSPRM5		0x0002
#define	BPC4_OSPRM4		0x0001


	/*
	 *		BEAMCON0 -- Beam Control Register
	 */

#define	BMC0_HARDDIS	0x4000	/* Disable hardware limits */
#define	BMC0_LPENDIS	0x2000	/* Disable light pen latch */
#define	BMC0_VARVBEN	0x1000	/* Enable variable vertical blank */
#define	BMC0_LOLDIS		0x0800	/* Disable long/short line toggle */
#define	BMC0_CSCBEN		0x0400	/* Composite sync/blank */
#define	BMC0_VARVSYEN	0x0200	/* Enable variable vertical sync */
#define	BMC0_VARHSYEN	0x0100	/* Enable variable horizontal sync */
#define	BMC0_VARBEAMEN	0x0080	/* Enable variable beam counters */
#define	BMC0_PAL			0x0020	/* Set decodes for PAL */
#define	BMC0_VARCSYEN	0x0010	/* Enable variable composite sync */
#define	BMC0_BLANKEN	0x0008	/* Blank enable? */
#define	BMC0_CSYTRUE	0x0004	/* CSY polarity */
#define	BMC0_VSYTRUE	0x0002	/* VSY polarity */
#define	BMC0_HSYTRUE	0x0001	/* HSY polarity */


	/*
	 *		FMODE -- Fetch Mode Control Register (AGA)
	 */

#define	FMODE_SSCAN2	0x8000	/* Sprite scan-doubling */
#define	FMODE_BSCAN2	0x4000	/* Use PF2 modulus every other line */
#define	FMODE_SPAGEM	0x0008	/* Sprite page mode */
#define	FMODE_SPR32		0x0004	/* Sprite 32 bit fetch */
#define	FMODE_BPAGEM	0x0002	/* Bitplane page mode */
#define	FMODE_BPL32		0x0001	/* Bitplane 32 bit fetch */


	/*
	 *		Pixel Clock Definitions
	 *
	 *		These depend on the E-Clock, so they are filled in dynamically
	 */

static u_long MasterClock;			/* 28 MHz Master Clock */
static u_long ColorClock;			/* 3.5 MHz Color Clock */

static u_long PixClock_LORES;		/* LORES, 140 ns /  7 MHz */
static u_long PixClock_HIRES;		/* HIRES,  70 ns / 14 MHz */
static u_long PixClock_SHRES;		/* SHRES,  35 ns / 28 MHz */


	/*
	 *    Maximum Display Depth
	 *
	 *    These depend on the Chipset, so they are filled in dynamically
	 */

static u_long MaxDepth_LORES;
static u_long MaxDepth_HIRES;
static u_long MaxDepth_SHRES;


	/*
	 *		Round Up/Down, Div/Mod Macros
	 */

#define up8(x)			(((x)+7)&~7)
#define down8(x)		((x)&~7)
#define div8(x)		((x)>>3)
#define mod8(x)		((x)&7)

#define up16(x)		(((x)+15)&~15)
#define down16(x)		((x)&~15)
#define div16(x)		((x)>>4)
#define mod16(x)		((x)&15)

#define up32(x)		(((x)+31)&~31)
#define down32(x)		((x)&~31)
#define div32(x)		((x)>>5)
#define mod32(x)		((x)&31)

#define up64(x)		(((x)+63)&~63)
#define down64(x)		((x)&~63)
#define div64(x)		((x)>>6)
#define mod64(x)		((x)&63)


	/*
	 *		High Word/Low Word Macros
	 */

#define highw(x)	((u_long)(x)>>16 & 0xffff) /* Get the High UWORD in a ULONG */
#define loww(x)	((u_long)(x) & 0xffff) 		/* Get the Low UWORD in a ULONG */


	/*
	 *		Chip RAM we reserve for the Frame Buffer
	 *
	 *		This defines the maximum virtual screen size
	 */

#define VIDEOMEMSIZE_AGA_2M	(1280*1024)		/* AGA (2M)	-> max 1280*1024*256 */
#define VIDEOMEMSIZE_AGA_1M	(1024*768)		/* AGA (1M)	-> max 1024*768*256 */
#define VIDEOMEMSIZE_ECS_2M	(1280*1024/2)	/* ECS (2M)	-> max 1280*1024*16 */
#define VIDEOMEMSIZE_ECS_1M	(1024*768/2)	/* ECS (1M)	-> max 1024*768*16 */
#define VIDEOMEMSIZE_OCS		(800*600/2)		/* OCS		-> max 800*600*16 */


static u_char *videomemory;
static u_long videomemorysize;

#define AssignChunk(name, type, ptr, size) \
{ \
	(name) = (type)(ptr); \
	ptr += size; \
}


	/*
	 *		Copper Instructions
	 */

#define CMOVE(val, reg)			((reg)<<16 | (val))
#define CWAIT(x, y)				(((y)&0xff)<<24 | ((x)&0xfe)<<16 | 0x0001fffe)
#define CEND						(0xfffffffe)


typedef union {
	u_long l;
	u_short w[2];
} copins;

struct CopDisplay {
	copins *Init;
	copins *LOFlist;
	copins *SHFlist;				/* Only used for Interlace */
};

static struct CopDisplay copdisplay[2];
static volatile u_long currentcop = 0;


	/*
	 *		Hardware Cursor
	 */

static u_long *LOFsprite, *SHFsprite, *dummysprite;

#define CRSR_RATE		24			/* Number of frames/flash toggle */


	/*
	 *		Split 24-bit RGB colors in 12-bit MSB (for OCS/ECS/AGA) and
	 *		LSB (for AGA)
	 */

#define COLOR_MSB(rgb)	(((rgb>>12)&0xf00)|((rgb>>8)&0x0f0)|((rgb>>4)&0x00f))
#define COLOR_LSB(rgb)	(((rgb>>8) &0xf00)|((rgb>>4)&0x0f0)|((rgb)   &0x00f))


	/*
	 *		External Font Definitions
	 */

extern unsigned char fontdata[];
extern int fontheight;
extern int fontwidth;


	/*
	 *		Current Video Mode
	 */

static struct amiga_fb_par {
	struct fb_var_screeninfo var;
	struct fb_fix_screeninfo fix;
	union {
		struct {
			u_short Palette[32];
		} OCS;
		struct {
			u_short Palette[32];
		} ECS;
		struct {
			u_long next_line;			/* Modulo for next line */
			u_long next_plane;		/* Modulo for next plane */
			u_short diwstrt_h;		/* vmode */
			u_short diwstop_h;		/* vmode */
			u_short diwstrt_v;		/* vmode */
			u_short diwstop_v;		/* vmode */
			u_short sprstrt_h;		/* vmode */
			u_short sprstrt_v;		/* vmode */
			long crsr_x;				/* movecursor */
			long crsr_y;				/* movecursor */
			u_long bplpt[8];			/* vmode, pan */
			u_long sprpt[2];			/* vmode */
			u_short bplcon0;			/* vmode */
			u_short bplcon1;			/* vmode, pan */
			u_short bplcon2;			/* vmode */
			u_short bplcon3;			/* vmode */
			u_short bplcon4;			/* vmode */
			u_short bpl1mod;			/* vmode, pan */
			u_short bpl2mod;			/* vmode, pan */
			u_short diwstrt;			/* vmode */
			u_short diwstop;			/* vmode */
			u_short diwhigh;			/* vmode */
			u_short ddfstrt;			/* vmode, pan */
			u_short ddfstop;			/* vmode, pan */
			u_short htotal;			/* vmode */
			u_short hsstrt;			/* vmode */
			u_short hsstop;			/* vmode */
			u_short hbstrt;			/* vmode */
			u_short hbstop;			/* vmode */
			u_short vtotal;			/* vmode */
			u_short vsstrt;			/* vmode */
			u_short vsstop;			/* vmode */
			u_short vbstrt;			/* vmode */
			u_short vbstop;			/* vmode */
			u_short hcenter;			/* vmode */
			u_short beamcon0;			/* vmode */
			u_short fmode;				/* vmode */
			u_long Palette[256];
		} AGA;
	} hw;
} current_par;

static int current_par_valid = 0; 
static int currcon = 0;

static struct display disp[NR_CONSOLES];
static struct fb_info fb_info;

static int node;			/* node of the /dev/fb?current file */


	/*
	 *		Latches for Display Changes during VBlank
	 */

static volatile u_short do_vmode = 0;				/* Change the video mode */
static volatile u_short do_movecursor = 0;		/* Move the cursor */
static volatile u_short do_blank = 0;				/* Blank the screen */
static volatile u_short do_unblank = 0;			/* Unblank the screen */


	/*
	 *		Various Flags
	 */

static volatile u_short full_vmode_change = 1;	/* Full change or only pan */
static volatile u_short is_blanked = 0;			/* Screen is blanked */



	/*
	 *		Switch for Chipset Independency
	 */

static struct fb_hwswitch {

	/* Initialisation */

	int (*init)(void);

	/* Display Control */

	int (*encode_fix)(struct fb_fix_screeninfo *fix, struct amiga_fb_par *par);
	int (*decode_var)(struct fb_var_screeninfo *var, struct amiga_fb_par *par);
	int (*encode_var)(struct fb_var_screeninfo *var, struct amiga_fb_par *par);
	void (*get_par)(struct amiga_fb_par *par);
	void (*set_par)(struct amiga_fb_par *par, u_short);
	int (*getcolreg)(u_int regno, u_int *red, u_int *green, u_int *blue,
						  u_int *transp);
	int (*setcolreg)(u_int regno, u_int red, u_int green, u_int blue,
						  u_int transp);

	/* Routines Called by VBlank Interrupt to Prevent Flicker */

	void (*do_vmode)(struct display *p);
	void (*do_movecursor)(struct display *p);
	void (*do_blank)(struct display *p);
	void (*do_unblank)(struct display *p);
	void (*do_flashcursor)(struct display *p);
} *fbhw;


	/*
	 *		Frame Buffer Names
	 */

static char amiga_AGA_fb_name[] = "Amiga AGA";
static char amiga_ECS_fb_name[] = "Amiga ECS";
static char amiga_OCS_fb_name[] = "Amiga OCS";
static char *amiga_fb_name = "Amiga Unknown";			/* Default */


	/*
	 *		Predefined Video Modes
	 *
	 *		Since the actual pixclock values depend on the E-Clock, we use the tags
	 *		defined below and fill in the real values during initialisation
	 */

char *amiga_fb_modenames[] = {
		"default",
		"pal-lace",
		"ntsc-lace",
		"vga",
		"vga70",
		"user1",
		"user2",
		"user3",
		"user4",
};


#define TAG_LORES			(1)		/* This assumes no one has pixel clocks of */
#define TAG_HIRES			(2)		/* 333, 500 or 1000 GHz :-) */
#define TAG_SHRES			(3)


struct fb_var_screeninfo amiga_fb_predefined[] = {

	/*
	 *		System Default Video Modes
	 */

	{
		/* Default */
		0, 0, 0, 0, 0, 0, 0, 0, 					  		/* xres-grayscale */
		{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, 	/* red green blue tran*/
		0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0
	}, {
		/* PAL lace, 15 kHz, 50 Hz */
		640, 512, 640, 512, 0, 0, 1, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCEL_NONE, TAG_HIRES, 146, 54, 80, 33, 68, 2,
		FB_SYNC_BROADCAST, FB_VMODE_INTERLACED
	}, {
		/* NTSC lace, 15 kHz, 60 Hz */
		640, 400, 640, 400, 0, 0, 1, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCEL_NONE, TAG_HIRES, 146, 54, 80, 33, 68, 2,
		FB_SYNC_BROADCAST, FB_VMODE_INTERLACED
	}, {
		/* VGA 31.5 kHz, 60 Hz */
		640, 480, 640, 480, 0, 0, 4, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCEL_NONE, TAG_SHRES, 64, 96, 30, 9, 112, 2,
		0, FB_VMODE_NONINTERLACED
	}, {
		/* VGA 31.5 kHz, 70 Hz */
		640, 400, 640, 400, 0, 0, 4, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCEL_NONE, TAG_SHRES, 64, 96, 35, 12, 112, 2,
		FB_SYNC_COMP_HIGH_ACT|FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED
	},

	/*
	 *		User Defined Video Modes: to be set after boot up using e.g. fbset
	 */

	{
		/* User 1 */
		0, 0, 0, 0, 0, 0, 0, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0
	}, {
		/* User 2 */
		0, 0, 0, 0, 0, 0, 0, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0
	}, {
		/* User 3 */
		0, 0, 0, 0, 0, 0, 0, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0
	}, {
		/* User 4 */
		0, 0, 0, 0, 0, 0, 0, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0
	}
};


int num_amiga_fb_predefined = sizeof(amiga_fb_predefined)/
										sizeof(struct fb_var_screeninfo);

static int amifb_ilbm = 0;			/* interleaved or normal bitplanes */
static int amifb_inverse = 0;
static int amifb_mode = -1;




/* Some default modes */

#define PAL_DEFMODE		1	/* PAL-laced for PAL OCS/ECS */
#define NTSC_DEFMODE		2	/* NTSC-laced for NTSC OCS/ECS */
#define AGA_DEFMODE		4	/* VGA70 for AGA */


	/*
	 *		Allocate, Clear and Align a Block of Chip Memory
	 */

static u_long chipalloc(u_long size)
{
	u_long ptr;

	size += PAGE_SIZE-1;
	if (!(ptr = (u_long)amiga_chip_alloc(size)))
		panic("No Chip RAM for frame buffer\n");
	memset((void *)ptr, 0, size);
	ptr = PAGE_ALIGN(ptr);

	return(ptr);
}


/* -------------------- OCS specific routines -------------------- */

#include "amifb_OCS.c"

/* -------------------- ECS specific routines -------------------- */

#include "amifb_ECS.c"

/* -------------------- AGA specific routines -------------------- */

#include "amifb_AGA.c"


/* -------------------- Interfaces to hardware functions -------------------- */


static struct fb_hwswitch OCS_switch = {
	OCS_init, OCS_encode_fix, OCS_decode_var, OCS_encode_var, OCS_get_par,
	OCS_set_par, OCS_getcolreg, OCS_setcolreg, OCS_do_vmode, OCS_do_movecursor,
	OCS_do_blank, OCS_do_unblank, OCS_do_flashcursor
};

static struct fb_hwswitch ECS_switch = {
	ECS_init, ECS_encode_fix, ECS_decode_var, ECS_encode_var, ECS_get_par,
	ECS_set_par, ECS_getcolreg, ECS_setcolreg, ECS_do_vmode, ECS_do_movecursor,
	ECS_do_blank, ECS_do_unblank, ECS_do_flashcursor
};

static struct fb_hwswitch AGA_switch = {
	AGA_init, AGA_encode_fix, AGA_decode_var, AGA_encode_var, AGA_get_par,
	AGA_set_par, AGA_getcolreg, AGA_setcolreg, AGA_do_vmode, AGA_do_movecursor,
	AGA_do_blank, AGA_do_unblank, AGA_do_flashcursor
};


static void amiga_fb_get_par(struct amiga_fb_par *par)
{
	if (current_par_valid)
		*par = current_par;
	else
		fbhw->get_par(par);
}


static void amiga_fb_set_par(struct amiga_fb_par *par)
{
	fbhw->set_par(par, 0);
	current_par = *par;
	current_par_valid = 1;
}


/* -------------------- Generic routines -------------------- */


static int do_fb_set_var(struct fb_var_screeninfo *var, int isactive)
{
	int err, activate;
	struct amiga_fb_par par;

	if ((err = fbhw->decode_var(var, &par)))
		return(err);
	activate = var->activate;
	if (((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) && isactive)
		amiga_fb_set_par(&par);
	fbhw->encode_var(var, &par);
	var->activate = activate;
	return(0);
}


	/*
	 *		Default Colormaps
	 */

static u_short red2[] =
	{ 0x0000, 0xc000};
static u_short green2[] =
	{ 0x0000, 0xc000};
static u_short blue2[] =
	{ 0x0000, 0xc000};

static u_short red4[] =
	{ 0x0000, 0xc000, 0x8000, 0xffff};
static u_short green4[] =
	{ 0x0000, 0xc000, 0x8000, 0xffff};
static u_short blue4[] =
	{ 0x0000, 0xc000, 0x8000, 0xffff};

static u_short red8[] =
	{ 0x0000, 0x0000, 0x0000, 0x0000, 0xc000, 0xc000, 0xc000, 0xc000};
static u_short green8[] =
	{ 0x0000, 0x0000, 0xc000, 0xc000, 0x0000, 0x0000, 0xc000, 0xc000};
static u_short blue8[] =
	{ 0x0000, 0xc000, 0x0000, 0xc000, 0x0000, 0xc000, 0x0000, 0xc000};

static u_short red16[] =
	{ 0x0000, 0x0000, 0x0000, 0x0000, 0xc000, 0xc000, 0xc000, 0xc000,
	  0x8000, 0x0000, 0x0000, 0x0000, 0xffff, 0xffff, 0xffff, 0xffff};
static u_short green16[] =
	{ 0x0000, 0x0000, 0xc000, 0xc000, 0x0000, 0x0000, 0xc000, 0xc000,
	  0x8000, 0x0000, 0xffff, 0xffff, 0x0000, 0x0000, 0xffff, 0xffff};
static u_short blue16[] =
	{ 0x0000, 0xc000, 0x0000, 0xc000, 0x0000, 0xc000, 0x0000, 0xc000,
	  0x8000, 0xffff, 0x0000, 0xffff, 0x0000, 0xffff, 0x0000, 0xffff};


static struct fb_cmap default_2_colors = { 0, 2, red2, green2, blue2, NULL };
static struct fb_cmap default_8_colors = { 0, 8, red8, green8, blue8, NULL };
static struct fb_cmap default_4_colors = { 0, 4, red4, green4, blue4, NULL };
static struct fb_cmap default_16_colors = { 0, 16, red16, green16, blue16,
														  NULL };


static struct fb_cmap *get_default_colormap(int bpp)
{
	switch (bpp) {
		case 1:
			return(&default_2_colors);
			break;
		case 2:
			return(&default_4_colors);
			break;
		case 3:
			return(&default_8_colors);
			break;
		default:
			return(&default_16_colors);
			break;
	}
}


#define CNVT_TOHW(val,width)		(((val)<<(width))+0x7fff-(val))>>16
#define CNVT_FROMHW(val,width)	((width) ? ((((val)<<16)-(val)) / \
															((1<<(width))-1)) : 0)


static int do_fb_get_cmap(struct fb_cmap *cmap, struct fb_var_screeninfo *var,
								  int kspc)
{
	int i, start;
	u_short *red, *green, *blue, *transp;
	u_int hred, hgreen, hblue, htransp;

	red = cmap->red;
	green = cmap->green;
	blue = cmap->blue;
	transp = cmap->transp;
	start = cmap->start;
	if (start < 0)
		return(-EINVAL);
	for (i = cmap->len; i--;) {
		if (fbhw->getcolreg(start++, &hred, &hgreen, &hblue, &htransp))
			return(0);
		hred = CNVT_FROMHW(hred, var->red.length);
		hgreen = CNVT_FROMHW(hgreen, var->green.length);
		hblue = CNVT_FROMHW(hblue, var->blue.length);
		htransp = CNVT_FROMHW(htransp, var->transp.length);
		if (kspc) {
			*red = hred;
			*green = hgreen;
			*blue = hblue;
			if (transp)
				*transp = htransp;
		} else {
			put_user_word(hred, red);
			put_user_word(hgreen, green);
			put_user_word(hblue, blue);
			if (transp)
				put_user_word(htransp, transp);
		}
		red++;
		green++;
		blue++;
		if (transp)
			transp++;
	}
	return(0);
}


static int do_fb_set_cmap(struct fb_cmap *cmap, struct fb_var_screeninfo *var,
								  int kspc)
{
	int i, start;
	u_short *red, *green, *blue, *transp;
	u_int hred, hgreen, hblue, htransp;

	red = cmap->red;
	green = cmap->green;
	blue = cmap->blue;
	transp = cmap->transp;
	start = cmap->start;

	if (start < 0)
		return(-EINVAL);
	for (i = cmap->len; i--;) {
		if (kspc) {
			hred = *red;
			hgreen = *green;
			hblue = *blue;
			htransp = transp ? *transp : 0;
		} else {
			hred = get_user_word(red);
			hgreen = get_user_word(green);
			hblue = get_user_word(blue);
			htransp = transp ? get_user_word(transp) : 0;
		}
		hred = CNVT_TOHW(hred, var->red.length);
		hgreen = CNVT_TOHW(hgreen, var->green.length);
		hblue = CNVT_TOHW(hblue, var->blue.length);
		htransp = CNVT_TOHW(htransp, var->transp.length);
		red++;
		green++;
		blue++;
		if (transp)
			transp++;
		if (fbhw->setcolreg(start++, hred, hgreen, hblue, htransp))
			return(0);
	}
	return(0);
}


static void do_install_cmap(int con)
{
	if (con != currcon)
		return;
	if (disp[con].cmap.len)
		do_fb_set_cmap(&disp[con].cmap, &disp[con].var, 1);
	else
		do_fb_set_cmap(get_default_colormap(disp[con].var.bits_per_pixel),
														&disp[con].var, 1);		
}


static void memcpy_fs(int fsfromto, void *to, void *from, int len)
{
	switch (fsfromto) {
		case 0:
			memcpy(to,from,len);
			return;
		case 1:
			memcpy_fromfs(to,from,len);
			return;
		case 2:
			memcpy_tofs(to,from,len);
			return;
	}
}


static void copy_cmap(struct fb_cmap *from, struct fb_cmap *to, int fsfromto)
{
	int size;
	int tooff = 0, fromoff = 0;

	if (to->start > from->start)
		fromoff = to->start-from->start;
	else
		tooff = from->start-to->start;			
	size = to->len-tooff;
	if (size > from->len-fromoff)
		size = from->len-fromoff;
	if (size < 0)
		return;
	size *= sizeof(u_short);
	memcpy_fs(fsfromto, to->red+tooff, from->red+fromoff, size);
	memcpy_fs(fsfromto, to->green+tooff, from->green+fromoff, size);
	memcpy_fs(fsfromto, to->blue+tooff, from->blue+fromoff, size);
	if (from->transp && to->transp)
		memcpy_fs(fsfromto, to->transp+tooff, from->transp+fromoff, size);
}


static int alloc_cmap(struct fb_cmap *cmap, int len, int transp)
{
	int size = len*sizeof(u_short);

	if (cmap->len != len) {
		if (cmap->red)
			kfree(cmap->red);
		if (cmap->green)
			kfree(cmap->green);
		if (cmap->blue)
			kfree(cmap->blue);
		if (cmap->transp)
			kfree(cmap->transp);
		cmap->red = cmap->green = cmap->blue = cmap->transp = NULL;
		cmap->len = 0;
		if (!len)
			return(0);
		if (!(cmap->red = kmalloc(size, GFP_ATOMIC)))
			return(-1);
		if (!(cmap->green = kmalloc(size, GFP_ATOMIC)))
			return(-1);
		if (!(cmap->blue = kmalloc(size, GFP_ATOMIC)))
			return(-1);
		if (transp) {
			if (!(cmap->transp = kmalloc(size, GFP_ATOMIC)))
				return(-1);
		} else
			cmap->transp = NULL;
	}
	cmap->start = 0;
	cmap->len = len;
	copy_cmap(get_default_colormap(len), cmap, 0);
	return(0);
}	


	/*
	 *		Get the Fixed Part of the Display
	 */

static int amiga_fb_get_fix(struct fb_fix_screeninfo *fix, int con)
{
	struct amiga_fb_par par;
	int error = 0;

	if (con == -1)
		amiga_fb_get_par(&par);
	else
		error = fbhw->decode_var(&disp[con].var, &par);
	return(error ? error : fbhw->encode_fix(fix, &par));
}

	
	/*
	 *		Get the User Defined Part of the Display
	 */

static int amiga_fb_get_var(struct fb_var_screeninfo *var, int con)
{
	struct amiga_fb_par par;
	int error = 0;

	if (con == -1) {
		amiga_fb_get_par(&par);
		error = fbhw->encode_var(var, &par);
	} else
		*var = disp[con].var;
	return(error);
}


static void amiga_fb_set_disp(int con)
{
	struct fb_fix_screeninfo fix;

	amiga_fb_get_fix(&fix, con);
	if (con == -1)
		con = 0;
	disp[con].screen_base = (u_char *)fix.smem_start;
	disp[con].type = fix.type;
	disp[con].type_aux = fix.type_aux;
	disp[con].ywrapstep = fix.ywrapstep;
	disp[con].fontdata = fontdata;
	disp[con].fontheight = fontheight;
	disp[con].fontwidth = fontwidth;
}


	/*
	 *		Set the User Defined Part of the Display
	 */

static int amiga_fb_set_var(struct fb_var_screeninfo *var, int con)
{
	int err, oldxres, oldyres, oldbpp;

	if ((err = do_fb_set_var(var, con == currcon)))
		return(err);
	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		oldxres = disp[con].var.xres;
		oldyres = disp[con].var.yres;
		oldbpp = disp[con].var.bits_per_pixel;
		disp[con].var = *var;
		if (oldxres != var->xres || oldyres != var->yres ||
			 oldbpp != var->bits_per_pixel) {
			amiga_fb_set_disp(con);
			(*fb_info.changevar)(con);
			alloc_cmap(&disp[con].cmap, 0, 0);
			do_install_cmap(con);
		}
	}
	var->activate = 0;
	return(0);
}


	/*
	 *		Get the Colormap
	 */

static int amiga_fb_get_cmap(struct fb_cmap *cmap, int kspc, int con)
{
	if (con == currcon) /* current console ? */
		return(do_fb_get_cmap(cmap, &disp[con].var, kspc));
	else if (disp[con].cmap.len) /* non default colormap ? */
		copy_cmap(&disp[con].cmap, cmap, kspc ? 0 : 2);
	else
		copy_cmap(get_default_colormap(disp[con].var.bits_per_pixel), cmap,
					 kspc ? 0 : 2);
	return(0);
}


	/*
	 *		Set the Colormap
	 */

static int amiga_fb_set_cmap(struct fb_cmap *cmap, int kspc, int con)
{
	int err;

	if (!disp[con].cmap.len) {			/* no colormap allocated ? */
		if ((err = alloc_cmap(&disp[con].cmap, 1<<disp[con].var.bits_per_pixel,
									 0)))
			return(err);
	}
	if (con == currcon)					/* current console ? */
		return(do_fb_set_cmap(cmap, &disp[con].var, kspc));
	else
		copy_cmap(cmap, &disp[con].cmap, kspc ? 0 : 1);
	return(0);
}


#if 0	/* Still need some discussion... */
	/*
	 *		Set the Hardware Cursor Image
	 */

static int amiga_fb_set_crsr(struct fb_cursor *crsr, int con)
{
	return(-EINVAL);
}


	/*
	 *		Move the Hardware Cursor
	 */

static int amiga_fb_move_crsr(int *offset, int con)
{
	return(-EINVAL);
}


	/*
	 *		Pan or Wrap the Display
	 */

static int amiga_fb_pan_display(int *offset, int con)
{
	return(-EINVAL);
}
#endif


	/*
	 *		Amiga Frame Buffer Specific ioctls
	 */

static int amiga_fb_ioctl(struct inode *inode, struct file *file,
								  u_int cmd, u_long arg, int con)
{
	return(-EINVAL);
}


static struct fb_ops amiga_fb_ops = {
	amiga_fb_get_fix, amiga_fb_get_var, amiga_fb_set_var, amiga_fb_get_cmap,
	amiga_fb_set_cmap,
#if 0
	amiga_fb_set_crsr, amiga_fb_move_crsr, amiga_fb_pan_display,
#endif
	amiga_fb_ioctl	
};


static int amifb_switch(int con)
{
	/* Do we have to save the colormap ? */
	if (disp[currcon].cmap.len)
		do_fb_get_cmap(&disp[currcon].cmap, &disp[currcon].var, 1);

	/* Reset hardware scrolling to make sure the display is continous */
   disp[con].var.yoffset = 0;
   disp[con].var.vmode &= ~FB_VMODE_WRAP;

	do_fb_set_var(&disp[con].var, 1);
	currcon = con;
	/* Install new colormap */
	do_install_cmap(con);
	return(0);
}


	/*
	 *		Update the `var' structure (called by amicon.c)
	 */

static int amifb_updatevar(int con)
{
	current_par.var.yoffset = disp[con].var.yoffset;
	current_par.var.vmode = disp[con].var.vmode;
	fbhw->set_par(&current_par, 1);
	return(0);
}


static void amifb_blank(int blank)
{
	if (blank) {
		do_unblank = 0;
		do_blank = 1;
	} else {
		do_blank = 0;
		do_unblank = 1;
	}
}


	/*
	 *		VBlank Display Interrupt
	 */

static void amifb_interrupt(struct intframe *fp, void *data)
{
   struct display *p = &disp[currcon];

	/* I *think* that you should only change display lists on long frame.
	 * At least it goes awfully perculiar on my A500 without the following
	 * test. Not really in a position to test this hypothesis, so sorry
	 * for the slow scrolling, all you flicker-fixed souls
	 */

	if (((p->var.vmode & FB_VMODE_MASK) != FB_VMODE_INTERLACED) ||
		 (custom.vposr & 0x8000)) {
		if (do_vmode) {
			fbhw->do_vmode(p);
			do_vmode = 0;
		}
		if (do_movecursor) {
			fbhw->do_movecursor(p);
			do_movecursor = 0;
		}
	}
	fbhw->do_flashcursor(p);
	if (do_blank) {
		fbhw->do_blank(p);
		do_blank = 0;
	}
	if (do_unblank) {
		fbhw->do_unblank(p);
		do_unblank = 0;
	}
}


void video_setup(char *options, int *ints)
{
	char *this_opt;
	int i;

	if (!options || !*options)
		return;

	for (this_opt = strtok(options, ","); this_opt; this_opt = strtok(NULL, ","))
		if (!strcmp(this_opt, "inverse")) {
			amifb_inverse = 1;
			for (i = 0; i < 16; i++) {
				red16[i] = ~red16[i];
				green16[i] = ~green16[i];
				blue16[i] = ~blue16[i];
			}
			for (i = 0; i < 8; i++) {
				red8[i] = ~red8[i];
				green8[i] = ~green8[i];
				blue8[i] = ~blue8[i];
			}
			for (i = 0; i < 4; i++) {
				red4[i] = ~red4[i];
				green4[i] = ~green4[i];
				blue4[i] = ~blue4[i];
			}
			for (i = 0; i < 2; i++) {
				red2[i] = ~red2[i];
				green2[i] = ~green2[i];
				blue2[i] = ~blue2[i];
			}
		} else if (!strcmp(this_opt, "ilbm"))
			amifb_ilbm = 1;
		else
			for (i = 0; i < num_amiga_fb_predefined; i++)
				if (!strcmp(this_opt, amiga_fb_modenames[i])) {
					amifb_mode = i;
					break;
				}
}


	/*
	 *		Initialisation
	 */

struct fb_info *amifb_init(long *mem_start)
{
	int err;
	struct amiga_fb_par par;
	u_long eclock = boot_info.bi_amiga.eclock;
	int i;

	switch (boot_info.bi_amiga.chipset) {
		case CS_AGA:
			amiga_fb_name = amiga_AGA_fb_name;
			fbhw = &AGA_switch;
			MaxDepth_LORES = 8;
			MaxDepth_HIRES = 8;
			MaxDepth_SHRES = 8;
			break;

		case CS_ECS:
			amiga_fb_name = amiga_ECS_fb_name;
			fbhw = &ECS_switch;
			MaxDepth_LORES = 6;
			MaxDepth_HIRES = 4;
			MaxDepth_SHRES = 2;
			break;

		case CS_OCS:
			amiga_fb_name = amiga_OCS_fb_name;
			fbhw = &OCS_switch;
			MaxDepth_LORES = 6;
			MaxDepth_HIRES = 4;
			MaxDepth_SHRES = 0;			/* OCS means no SHRES? */
			break;

		default:
			printk("Unknown graphics chipset, defaulting to OCS\n");
			fbhw = &OCS_switch;
			MaxDepth_LORES = 0;			/* 0 is always safe here */
			MaxDepth_HIRES = 0;
			MaxDepth_SHRES = 0;
			break;
	}

	err = register_framebuffer(amiga_fb_name, &node, &amiga_fb_ops,
										num_amiga_fb_predefined, amiga_fb_predefined);
	if (err < 0)
		panic("Cannot register frame buffer\n");

	/*
	 *		Calculate the Pixel Clock Values for this Machine
	 */

	MasterClock = 40*eclock;			/* 28 MHz Master Clock */
	ColorClock = 5*eclock;				/* 3.5 MHz Color Clock */
	PixClock_SHRES =  25000000/(eclock/1000);	/* SHRES,  35 ns / 28 MHz */
	PixClock_HIRES =  50000000/(eclock/1000);	/* HIRES,  70 ns / 14 MHz */
	PixClock_LORES = 100000000/(eclock/1000);	/* LORES, 140 ns /  7 MHz */

	/*
	 *		Replace the Tag Values with the Real Pixel Clock Values
	 */

	for (i = 0; i < num_amiga_fb_predefined; i++)
		switch (amiga_fb_predefined[i].pixclock) {
			case TAG_LORES:
				amiga_fb_predefined[i].pixclock = PixClock_LORES;
  				if (amiga_fb_predefined[i].bits_per_pixel > MaxDepth_LORES)
  					amiga_fb_predefined[i].bits_per_pixel = MaxDepth_LORES;
				break;

			case TAG_HIRES:
				amiga_fb_predefined[i].pixclock = PixClock_HIRES;
  				if (amiga_fb_predefined[i].bits_per_pixel > MaxDepth_HIRES)
  					amiga_fb_predefined[i].bits_per_pixel = MaxDepth_HIRES;
				break;

			case TAG_SHRES:
				amiga_fb_predefined[i].pixclock = PixClock_SHRES;
  				if (amiga_fb_predefined[i].bits_per_pixel > MaxDepth_SHRES)
  					amiga_fb_predefined[i].bits_per_pixel = MaxDepth_SHRES;
				break;

			default:
				/* Other pixclock value: don't touch */
				break;
		}

	fbhw->init();
	fbhw->get_par(&par);
	fbhw->encode_var(&amiga_fb_predefined[0], &par);

	if (!mach_add_isr(IRQ_AMIGA_VERTB, (isrfunc)amifb_interrupt, 0, NULL))
		panic("Couldn't add vblank interrupt");

	strcpy(fb_info.modename, amiga_fb_name);
	strcat(fb_info.modename, " ");
	strcat(fb_info.modename, amiga_fb_modenames[amifb_mode]);
	fb_info.disp = disp;
	fb_info.switch_con = &amifb_switch;
	fb_info.updatevar = &amifb_updatevar;
	fb_info.blank = &amifb_blank;

	do_fb_set_var(&par.var, 1);
	amiga_fb_get_var(&disp[0].var, -1);
	amiga_fb_set_disp(-1);
	do_install_cmap(0);
	return(&fb_info);
}
